/*  OSC Controller with value and text feedback
 *  Can receive messages and or bundles.
 *  Works out of the box with the default Reaper OSC configuration, or at least it should :-)
 *  The controller settings (number of tracks, insets, insert parameters) are sent to Reaper on startup. 
 *  'Reset all control surfaces' action command is sent on startup to synchronize state
 * 
 *  Track messages (bundles) are parsed and the state of the tracks is stored in structs.
 *  The selected track and the parameters of the selected plugin are stored seperately
 *  
 *  The design maintains state on the controller side for the track and plugin data.
 *  All required track and plugin data are retrieved and bank switching is handled on the controller itself.

 *  
 *  All included libraries are part of the default Teensyduino installation. 
 *  
 *  Hardware used:
 *  WIZ820_SD_ADAPTOR, https://www.pjrc.com/store/wiz820_sd_adaptor.html
 *  WIZ850io module. Connected as described on the WIZ820_SD adaptor page.
 *   
 *  TODO
 *  - remember selected fx parameter bank when switching between inserts
 */

#include <Bounce.h>
#include <Encoder.h>
#include <ResponsiveAnalogRead.h>
#include <SPI.h>
//#include <Ethernet.h>
#include <EthernetUdp.h>
#include <OSCBundle.h>
#include "Reaper.h"
#include "PluginMap.h"

//initialise Reaper object with track model
Reaper reaper;

//***** DISPLAY *****
// Constants for display. These are copied fronm the 4DSystems genieArduino library, just the ones that are needed.
// The library provided by 4DSystems is NOT required.
// Parts of the library have been copied and adapted to work asynchronous as the 4D library is blocking program execution while waiting for acknoledgement from the display.
#define GENIE_ACK             0x06
#define GENIE_NAK             0x15
#define TIMEOUT_PERIOD        1000
#define RESYNC_PERIOD         100
#define GENIE_WRITE_OBJ       1
#define GENIE_WRITE_STR       2
#define GENIE_WRITE_CONTRAST  4
#define GENIE_OBJ_KNOB        1
#define GENIE_OBJ_USERIMAGES  27
#define GENIE_OBJ_STRINGS     17
#define GENIE_OBJ_GAUGE       11

#define GENIE_LINK_IDLE       0
#define GENIE_LINK_WFAN       1     // waiting for Ack or Nak

#define ERROR_NONE            0
#define ERROR_NAK             -4    // 252  0xFC

#define RESETLINE             11    // Pin used for display reset on startup, currently not used
#define NUM_OF_SCREEN_OBJECTS 31     // number of 4D genie objects on the display, used for queue handling


// Screen objects
// string 0-3       value labels under knobs
// string 4-7       values for knobs
// image  8-11      virtual knobs
// string 8-11      labels above knobs
// string 12        track number
// string 15        track name
// string 13        song position
// string 14        tempo
// gauge  0         VU L
// gauge  1         VU R
// image  0-3       Virtual button above knobs 0=blank, 1=green, 2=bypass, 3=mute
// image  4-6       Devider lines between knobs 0=blank, 1=half, 2=full
// image  7         Plugin icons
// string 16        big label
HardwareSerial  displaySerial                             =Serial1;             // serial port for display
int             genieState                                = GENIE_LINK_IDLE;    // state is idle on startup
int             objectQueue[NUM_OF_SCREEN_OBJECTS+1];                           // screen objects  0, string 0-7, knob 0-3,gauge 0-1,
int             queueFirstIndex                           = 0;                  // object array index of first in queue, 0 if empty
int             queueLastNo                               = 1;                  // sequence number of last in queue
elapsedMicros   sinceDisplayQueueUpdate;                                        // timer for updating the display queue
unsigned int    displayUpdateInterval                     = 1100;               // display queue update interval in microseconds
//***** END DISPLAY *****


// ***** ENCODERS *****
#define NUM_OF_ENCODERS   6                                                     // constant for number of encoders for use in loops
Encoder       encoder[NUM_OF_ENCODERS]              = {Encoder(28,27),Encoder(26,25),Encoder(24,8),Encoder(7,6),Encoder(5,4),Encoder(3,2)};
long          currentEncoderValue[NUM_OF_ENCODERS]  = {0,0,0,0,0,0};            // array for encoder values
int           parameterValue[NUM_OF_ENCODERS]       = {0,0,0,0,0,0};            // array for parameter values
int           parameterDisplay[NUM_OF_ENCODERS]     = {0,0,0,0,0,0};            // array for scaled display values
int           parameterMin[NUM_OF_ENCODERS]         = {0,0,0,0,0,0};            // arrays for scaling and constraint purposes
int           parameterMax[NUM_OF_ENCODERS]         = {212,212,212,212,212,212};
elapsedMicros sinceEncoderCheck;                                                // timer for rotary encoder check
unsigned int  encoderCheckInterval                  = 900;                      // interval in microseconds for checking rotary encoders

// ***** END  ENCODERS *****



// ***** BUTTONS *****
// The buttons are handled in two groups: Panel button and Encoder buttons.
// The panel buttons fit in a unsigned long and the encoder buttons fit in a unsigned int, the same timer is used for debouncing both groups of buttons.
#define       CLOCK_PIN                   18                      // Connects to the Clock pin (CLK) of the 74HC165
#define       PLOAD_PIN                   19                      // Connects to Parallel load pin (SH/LD) of the 74HC165
#define       DATA_PIN                    20                      // Connects to the Serial out pin (Qh) of the 74HC165
#define       CLOCK_PIN_ENCODERS          17                      // encoder switch Clock 
#define       PLOAD_PIN_ENCODERS          16                      // encoder switch Parallel load
#define       DATA_PIN_ENCODERS           15                      // encoder switch Serial out
#define       NUM_OF_BUTTONS              8                       // number of button on panel
#define       NUM_OF_ENCODER_BUTTONS      8                       // number of encoder switches, minimum is 8
unsigned int  newPanelButtonStates;                               // most recently read states
unsigned int  currentPanelButtonStates;                           // actual states
unsigned int  debouncePanelButtonStates   = 0;                    // states at beginning of debounce interval
unsigned int  newEncoderButtonStates;                             // most recently read states
unsigned int  currentEncoderButtonStates;                         // actual states
unsigned int  debounceEncoderButtonStates = 0;                    // states at beginning of debounce interval
elapsedMicros sinceButtonCheck;                                   // timer for button check
unsigned int  buttonCheckInterval         = 1000;                 // interval in microseconds for checking buttons
elapsedMillis debounceTimer;                                      // timer for button debounce
elapsedMillis debounceEncoderTimer;                               // timer for encoder button debounce
unsigned int  debounceDelay               = 5;                    // the debounce time in milliseconds
boolean       currentButtonState[NUM_OF_BUTTONS];                 // boolean array for button states, N.B. pressed = LOW, released = HIGH
boolean       currentEncoderButtonState[NUM_OF_ENCODER_BUTTONS];  // boolean array for encoder button states, N.B. pressed = LOW, released = HIGH


enum panelButtons{  
                    B_CURSOR_UP,
                    B_CURSOR_DOWN,
                    B_CURSOR_LEFT,
                    B_CURSOR_RIGHT,
                    B_TRACK,
                    B_INSERTS,
                    B_PLUGIN,
                    B_MASTER};

// Bounce objects for the footswitch
int           debounceTime=5;
Bounce        footSwitch1(14,debounceTime);
Bounce        footSwitch2(39,debounceTime);



// ***** END BUTTONS *****

// ***** LEDS *****
// NB CLOCK & LATCH currently switched!
#define       LATCH_PIN                   21            // Pin connected to ST_CP (pin 12) of 74HC595
#define       LED_CLOCK_PIN               22            // Pin connected to SH_CP (pin 11) of 74HC595
#define       LED_DATA_PIN                23            // Pin connected to DS (pin 14) of 74HC595
boolean       refreshLeds                 = false;      // boolean indicating if leds need to be refreshed.
boolean       panelLedState[NUM_OF_BUTTONS];                         // boolean array for LED states                 
int           panelLedParameterMap[NUM_OF_BUTTONS] ={0,0,0,0,0,1,2,3};  
// ***** END LEDS *****




// ***** PEDAL *****
int           pedalValue;                             // analog read value of pedal
int           pedalParameterAddress[3]={1,1,1};       // array with track, fx and parameter number
ResponsiveAnalogRead pedal(A10,true);
// ***** END PEDAL *****

// you can find this written on the board of some Arduino Ethernets or shields; I couldn't find it but it does not matter :-)
byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED }; 

//ports
const unsigned int serverPort  = 8000;     // Device port on Reaper control panel
const unsigned int destPort    = 9000;     // Listen port on Reaper control panel

//Create UDP message object
EthernetUDP Udp;

//the Teensy's IP
IPAddress ip(192, 168, 178, 177);
//destination IP
IPAddress outIp(192,168,178,10);

// General Controller
# define NUM_OF_CONTROLS 4
enum controlModes{  
                    VOL_PAN,
                    SENDS,
                    MASTER,
                    EQ_SHELF,
                    EQ_BAND,
                    INSERTS,
                    PLUGIN};
int   selectedMode;               // selected controller mode
char  emptyLabel[19]=      {"                  "};
char  sendsLabel[19]=      {"      SENDS       "};
char  volumeLabel[19]=     {"      VOLUME      "};
char  panLabel[19]=        {"       PAN        "};
char  mixLabel[19]=        {"        MIX       "};
char  tempoLabel[19]=      {"      TEMPO       "};
char  zeroLabel[19]=       {"       0.0        "};
char  zeroString[7]=       {"0.0000"};
char  knobLabel[4][19];
char  centeredString[18]={};
int   encoderButtonIndex[8]={0,0,3,2,1,0,0,0};
int   deviderImageIndex[3]={0,0,0};                                           // array containing index of devider image between knobs
int   sourceTrackNumber=1;                                                    // used to store source track number when jumping to send track
int   selectedInsertsPage=0;                                                  // for scrolling through inserts
float encoderResolution=1.0/212.0;                                            // normalised encoder resolution
int   pluginId=0;                                                             // id = array index of identified plugin
// timer for ignoring parameter value feedback when turning rotary encoders
elapsedMillis feedbackIgnoreTimer[NUM_OF_CONTROLS];                           // timer for encoder feedback block
unsigned int  feedbackIgnoreDelay                       = 100;                // the time feedback from Reaper is ignored in milliseconds
boolean       processFeedback[NUM_OF_CONTROLS];
int           parameterOffset=0;                                              // parameter offset for selected page
int           parameterCount=0;                                               // number of parameters in selected plugin


elapsedMicros   functionTimer;
unsigned int    functionEndTime;



void setup(){
  Serial.begin(115200); 
  for (int i=0;i<NUM_OF_CONTROLS;i++){
    processFeedback[i]= true;
  }

    //****BUTTONS****
  // init button state arrays
  for (int i=0;i<NUM_OF_BUTTONS;i++){
    currentButtonState[i]=LOW;
  }
  for (int i=0;i<NUM_OF_ENCODER_BUTTONS;i++){
    currentEncoderButtonState[i]=LOW;
  }
  pinMode(PLOAD_PIN, OUTPUT);
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(DATA_PIN, INPUT);
  digitalWrite(CLOCK_PIN, LOW);
  digitalWrite(PLOAD_PIN, HIGH);
  pinMode(PLOAD_PIN_ENCODERS, OUTPUT);
  pinMode(CLOCK_PIN_ENCODERS, OUTPUT);
  pinMode(DATA_PIN_ENCODERS, INPUT);
  digitalWrite(CLOCK_PIN_ENCODERS, LOW);
  digitalWrite(PLOAD_PIN_ENCODERS, HIGH);
  // footswitch buttons
  pinMode(14, INPUT_PULLUP);
  pinMode(39, INPUT_PULLUP);


  //****LEDS****
  pinMode(LATCH_PIN, OUTPUT);
  pinMode(LED_DATA_PIN, OUTPUT);  
  pinMode(LED_CLOCK_PIN, OUTPUT);
  // initialize ledState array 
  for (int i=0;i<NUM_OF_BUTTONS;i++) panelLedState[i]=LOW;
  updateLeds();
  
  //setup ethernet part
  Ethernet.begin(mac,ip);
  Udp.begin(serverPort);
  
  delay(1000);
  
  // init display
  displaySerial.begin(200000);        // Serial1 @ 200000 (200K) Baud
  delay(1000);
  genieWriteContrast(14); 

  // send bundle with controller definition & state
  OSCBundle bndl;
  bndl.add("/device/track/count").add(NUM_OF_TRACKS);
  bndl.add("/device/send/count").add(NUM_OF_SENDS);
  bndl.add("/device/fx/count").add(NUM_OF_INSERTS);
  bndl.add("/device/fxparam/count").add(NUM_OF_PARAMETERS);
  bndl.add("/device/track/select").add(reaper.selectedTrack.number);
  bndl.add("/device/fx/select").add(reaper.selectedInsert.number);
  bndl.add("/device/fxparam/bank/select").add(reaper.selectedParameterBank);

  Udp.beginPacket(outIp, destPort);
  bndl.send(Udp);
  Udp.endPacket(); // mark the end of the OSC Packet
  bndl.empty(); // empty the bundle to free room for a new one

  selectMode(VOL_PAN);

  // send 'reset all controllers' message so Reaper will send all track data
  sendOscMessage("/action",41743);

}

void loop(){
  //process received messages
  oscReceive();
  // check elapsedMicros and millis timers
  if (sinceButtonCheck >=buttonCheckInterval) {
    sinceButtonCheck = sinceButtonCheck - buttonCheckInterval;
    checkButtons();
    checkEncoders();
    readPedal();
  }  
  if (sinceDisplayQueueUpdate >=displayUpdateInterval) {
    sinceDisplayQueueUpdate = sinceDisplayQueueUpdate - displayUpdateInterval;
    updateDisplay();  
    if (refreshLeds){
      updateLeds();
    }
  } 
  for (int i=0;i<NUM_OF_CONTROLS;i++){
    if (!processFeedback[i]){
      if (feedbackIgnoreTimer[i]>= feedbackIgnoreDelay) processFeedback[i]=true;
    }
  }

}

void oscReceive(){
  int size;
  if( (size = Udp.parsePacket())>0) {
    byte udpData[size];
    for (int i=0;i<size;i++) udpData[i]=Udp.read();
    // if data begins with # it is a bundle
    if (udpData[0]==35){
      OSCBundle   bundleIN;
      bundleIN.fill(udpData,size);
      if(!bundleIN.hasError()) {
        //Serial.println("Receiving OSC Bundle...");
        bundleIN.route("/track", parseTrackMessage);
        bundleIN.route("/fx", parseInsertMessage);
        bundleIN.route("/fxparam", parseInsertParameterMessage);
        //bundleIN.route("/fxeq", parseEqMessage);
        //bundleIN.route("/fxparam", showMessage);
        //bundleIN.route("/track", showMessage);
        //bundleIN.route("/*", showMessage);
        //bundleIN.route("/fxinstparam", showMessage);
        bundleIN.route("/beat",parseGlobalMessage);
        //bundleIN.route("/beat",showMessage);
        bundleIN.route("/tempo",parseGlobalMessage);
        bundleIN.route("/master",parseMasterMessage);
      }
    }
    // if data begins with / it is a message
    else if (udpData[0]==47){
      OSCMessage  msgIN;
      msgIN.fill(udpData,size);
      if(!msgIN.hasError()){
        //Serial.println("Receiving OSC Message...");
        msgIN.route("/track",parseTrackMessage);
        //msgIN.route("/fx",showMessage);
        //msgIN.route("/fxeq",showMessage);
        //msgIN.route("/fxparam",showMessage);
        msgIN.route("/beat",parseGlobalMessage);
        //msgIN.route("/*",showMessage);
      }
    }
  }
}

// send osc message, float
void sendOscMessage(const char* address, float value){
  OSCMessage msgOUT(address);         // set address
  msgOUT.add(value);                  // set value
  Udp.beginPacket(outIp, destPort);   // begin transmission
  msgOUT.send(Udp);                   // send the bytes
  Udp.endPacket();                    // mark the end of the OSC Packet
  msgOUT.empty();                     // free space occupied by message
}

// send osc message, int
void sendOscMessage(const char* address, int value){
  OSCMessage msgOUT(address);         // set address
  msgOUT.add(value);                  // set value
  Udp.beginPacket(outIp, destPort);   // begin transmission
  msgOUT.send(Udp);                   // send the bytes
  Udp.endPacket();                    // mark the end of the OSC Packet
  msgOUT.empty();                     // free space occupied by message
}

// show message address and contend on serial monitor
void showMessage(OSCMessage &msg, int addrOffset){
  char adresStr[48];
  msg.getAddress(adresStr);
  Serial.print(adresStr);
  Serial.print("\t\t");
  if (msg.isInt(0)){        
    Serial.print("Int     = : ");
    Serial.println(msg.getInt(0));
  } 
  else if(msg.isFloat(0)){       
    Serial.print("Float   = : ");
    Serial.println(msg.getFloat(0));
  } 
  else if(msg.isBoolean(0)){       
    Serial.print("Boolean = : ");
    Serial.println(msg.isBoolean(0));
  }
  else if(msg.isString(0)){   
    char str[48]    ;
    msg.getString(0,str,48);
    Serial.print("String  = : ");
    Serial.println(str);
  }
}

// parse global message
void parseGlobalMessage(OSCMessage &msg, int addrOffset ){
  if (msg.match("/beat/str")){
    msg.getString(0,reaper.beatStr,NAME_LENGTH); 
    char* beatRest=reaper.beatStr;
    // get measure part from beat string
    char* path=strtok_r(beatRest,".",&beatRest);
    int newMeasure=atoi(path);
    if (newMeasure!=reaper.measure) addToQueue(19);
    reaper.measure=newMeasure;
    // get beat part from beat string
    path=strtok_r(beatRest,".",&beatRest);
    int newBeat=atoi(path);
    // get fraction part from beat string
    path=strtok_r(beatRest,".",&beatRest);
    reaper.fraction=atoi(path);
    if (newBeat!=reaper.beat) addToQueue(19);
    reaper.beat=newBeat; 
    return;
  } 
  if (msg.match("/tempo/raw")){
    if (selectedMode==MASTER and processFeedback[0]){
      reaper.tempo=msg.getFloat(0); 
      addToQueue(20);
      addToQueue(9);
      return;
    }
    if (selectedMode!=MASTER){
      reaper.tempo=msg.getFloat(0); 
      addToQueue(20);
      return;
    }
    return;
  } 
}

// parse master message
void parseMasterMessage(OSCMessage &msg, int addrOffset ){
  if (msg.match("/vu/L",addrOffset)){
    reaper.masterVuL=msg.getFloat(0); 
    if (selectedMode==MASTER) addToQueue(21);
    return;
  }   
  if (msg.match("/vu/R",addrOffset)){
    reaper.masterVuR=msg.getFloat(0); 
    if (selectedMode==MASTER) addToQueue(22);
    return;
  }    
  if (msg.match("/volume/str",addrOffset)){
    msg.getString(0,reaper.masterVolumeStr,VALUE_LENGTH); 
    if (selectedMode==MASTER) addToQueue(8);
    return;
  }
  if (msg.match("/volume",addrOffset)){
    reaper.masterVolume=msg.getFloat(0); 
    //Serial.println(reaper.masterVolume);
    if (selectedMode==MASTER) addToQueue(12);
    return;
  }
  if (msg.match("/pan/str",addrOffset)){
    msg.getString(0,reaper.masterPanStr,VALUE_LENGTH); 
    if (selectedMode==MASTER) addToQueue(7);
    return;
  }
  if (msg.match("/pan",addrOffset)){
    reaper.masterPan=msg.getFloat(0); 
    if (selectedMode==MASTER) addToQueue(11);
    return;
  }    
  // master sends
  if (msg.match("/send",addrOffset)){
    char adresStr[48];    
    msg.getAddress(adresStr);
    int sendIndex = getAddresPartAsIndex(adresStr,3);
    if (sendIndex>=0){
      if (msg.match("/send/*/name",addrOffset)){
        msg.getString(0,reaper.masterSend[sendIndex].name,NAME_LENGTH); 
        //if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+1);
        return;
      }
      if (msg.fullMatch("/send/*/volume/str",addrOffset)){
        msg.getString(0,reaper.masterSend[sendIndex].volumeStr,VALUE_LENGTH); 
        //if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+5);
        return;
      }
      if (msg.fullMatch("/send/*/volume",addrOffset)){
        reaper.masterSend[sendIndex].volume=msg.getFloat(0); 
        //if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+9);
        return;
      }
      if (msg.fullMatch("/send/*/pan/str",addrOffset)){
        msg.getString(0,reaper.masterSend[sendIndex].panStr,VALUE_LENGTH); 
        return;
      }
      if (msg.fullMatch("/send/*/pan",addrOffset)){
        reaper.masterSend[sendIndex].pan=msg.getFloat(0); 
        return;
      }
    }

  }
}

// parse track message and store incoming data in structs
void parseTrackMessage(OSCMessage &msg, int addrOffset ){
  char adresStr[48];    
  msg.getAddress(adresStr);
  int trackIndex=getAddresPartAsIndex(adresStr,2);
  // numbered track data
  if (trackIndex>=0 and trackIndex<NUM_OF_TRACKS){
    if (msg.match("/*/number",addrOffset)){
      char number[2];
      msg.getString(0,number,2);
      reaper.track[trackIndex].number=atoi(number);
      return;
    }
    if (msg.match("/*/name",addrOffset)){
      msg.getString(0,reaper.track[trackIndex].name,NAME_LENGTH); 
      return;
    }
    if (msg.fullMatch("/*/volume/str",addrOffset)){
      msg.getString(0,reaper.track[trackIndex].volumeStr,VALUE_LENGTH); 
      return;
    }
    if (msg.fullMatch("/*/volume",addrOffset)){
      reaper.track[trackIndex].volume=msg.getFloat(0); 
      return;
    }
    if (msg.fullMatch("/*/pan/str",addrOffset)){
      msg.getString(0,reaper.track[trackIndex].panStr,VALUE_LENGTH); 
      return;
    }
    if (msg.fullMatch("/*/pan",addrOffset)){
      reaper.track[trackIndex].pan=msg.getFloat(0); 
      return;
    }
    if (msg.fullMatch("/*/recarm",addrOffset)){
      reaper.track[trackIndex].rec=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/*/mute",addrOffset)){
      reaper.track[trackIndex].mute=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/*/solo",addrOffset)){
      reaper.track[trackIndex].solo=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/*/monitor",addrOffset)){
      reaper.track[trackIndex].monitor=int(msg.getFloat(0)); 
      return;
    }

    // track sends
    if (msg.match("/*/send",addrOffset)){
      msg.getAddress(adresStr);
      int sendIndex = getAddresPartAsIndex(adresStr,4);
      if (sendIndex>=0){
        if (msg.match("/*/send/*/name",addrOffset)){
          msg.getString(0,reaper.track[trackIndex].send[sendIndex].name,NAME_LENGTH); 
          return;
        }
        if (msg.fullMatch("/send/*/volume/str",addrOffset)){
          msg.getString(0,reaper.track[trackIndex].send[sendIndex].volumeStr,VALUE_LENGTH); 
          return;
        }
        if (msg.fullMatch("/*/send/*/volume",addrOffset)){
          reaper.track[trackIndex].send[sendIndex].volume=msg.getFloat(0); 
          return;
        }
        if (msg.fullMatch("/*/send/*/pan/str",addrOffset)){
          msg.getString(0,reaper.track[trackIndex].send[sendIndex].panStr,VALUE_LENGTH); 
          return;
        }
        if (msg.fullMatch("/*/send/*/pan",addrOffset)){
          reaper.track[trackIndex].send[sendIndex].pan=msg.getFloat(0); 
          return;
        }
      }

    }
    // selected track inserts
    if (msg.match("/*/fx",addrOffset)){
      msg.getAddress(adresStr);
      int insertIndex = getAddresPartAsIndex(adresStr,4);
      if (insertIndex>=0){
        if (msg.match("/*/fx/*/name",addrOffset)){
          msg.getString(0,reaper.track[trackIndex].insert[insertIndex].name,NAME_LENGTH); 
          return;
        }
        if (msg.match("/*/fx/*/bypass")){
          reaper.track[trackIndex].insert[insertIndex].bypass=int(msg.getFloat(0)); 
          return;
        }
      }

    }
    //showMessage(msg,addrOffset);
  }
  // selected track data
  if (trackIndex<0){    
    if (msg.match("/vu/L",addrOffset)){
      reaper.trackVuL=msg.getFloat(0); 
      if (selectedMode!=MASTER) addToQueue(21);
      return;
    }   
    if (msg.match("/vu/R",addrOffset)){
      reaper.trackVuR=msg.getFloat(0); 
      if (selectedMode!=MASTER) addToQueue(22);
      return;
    }
    if (msg.match("/number",addrOffset)){
          char number[2];
          msg.getString(0,number,2);
          reaper.selectedTrack.number=atoi(number);
          addToQueue(17);
          return;
    }
    if (msg.match("/name",addrOffset)){
      msg.getString(0,reaper.selectedTrack.name,NAME_LENGTH); 
      addToQueue(18);
      return;
    }
    if (msg.fullMatch("/volume/str",addrOffset)){
      msg.getString(0,reaper.selectedTrack.volumeStr,VALUE_LENGTH); 
      if (selectedMode==VOL_PAN) addToQueue(8);
      return;
    }
    if (msg.fullMatch("/volume",addrOffset)){
      reaper.selectedTrack.volume=msg.getFloat(0); 
      if (selectedMode==VOL_PAN) addToQueue(12);
      return;
    }
    if (msg.fullMatch("/pan/str",addrOffset)){
      msg.getString(0,reaper.selectedTrack.panStr,VALUE_LENGTH); 
      if (selectedMode==VOL_PAN) addToQueue(7);
      return;
    }
    if (msg.fullMatch("/pan",addrOffset)){
      reaper.selectedTrack.pan=msg.getFloat(0); 
      if (selectedMode==VOL_PAN) addToQueue(11);
      return;
    }
    if (msg.fullMatch("/recarm",addrOffset)){
      reaper.selectedTrack.rec=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/mute",addrOffset)){
      reaper.selectedTrack.mute=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/solo",addrOffset)){
      reaper.selectedTrack.solo=int(msg.getFloat(0)); 
      return;
    }
    if (msg.fullMatch("/monitor",addrOffset)){
      reaper.selectedTrack.monitor=int(msg.getFloat(0)); 
      return;
    }
    // selected track sends
    if (msg.match("/send",addrOffset)){
      char adresStr[48];    
      msg.getAddress(adresStr);
      int sendIndex = getAddresPartAsIndex(adresStr,3);
      if (sendIndex>=0){
        if (msg.match("/send/*/name",addrOffset)){
          msg.getString(0,reaper.selectedTrack.send[sendIndex].name,18); 
          if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+1);
          return;
        }
        if (msg.fullMatch("/send/*/volume/str",addrOffset)){
          msg.getString(0,reaper.selectedTrack.send[sendIndex].volumeStr,VALUE_LENGTH); 
          if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+5);
          return;
        }
        if (msg.fullMatch("/send/*/volume",addrOffset)){
          reaper.selectedTrack.send[sendIndex].volume=msg.getFloat(0); 
          if (selectedMode==VOL_PAN and sendIndex<2) addToQueue(sendIndex+9);
          return;
        }
        if (msg.fullMatch("/send/*/pan/str",addrOffset)){
          msg.getString(0,reaper.selectedTrack.send[sendIndex].panStr,VALUE_LENGTH); 
          return;
        }
        if (msg.fullMatch("/send/*/pan",addrOffset)){
          reaper.selectedTrack.send[sendIndex].pan=msg.getFloat(0); 
          return;
        }
      }

    }
    // selected track inserts
    if (msg.match("/fx",addrOffset)){
      //Serial.println("Insert");
      char adresStr[48];    
      msg.getAddress(adresStr);
      int insertIndex = getAddresPartAsIndex(adresStr,3);
      if (insertIndex>=0){
        if (msg.match("/fx/*/name",addrOffset)){
          msg.getString(0,reaper.selectedTrack.insert[insertIndex].name,NAME_LENGTH); 
          if(selectedMode==INSERTS) {
            if (insertIndex<NUM_OF_CONTROLS and selectedInsertsPage==0) addToQueue(insertIndex+1);
            if (insertIndex>=NUM_OF_CONTROLS and selectedInsertsPage==1) addToQueue(insertIndex-NUM_OF_CONTROLS+1);
          }
          return;
        }
        if (msg.match("/fx/*/wetdry",addrOffset)){
          reaper.selectedTrack.insert[insertIndex].mix=msg.getFloat(0); 
          return;
        }
        if (msg.match("/fx/*/wetdry/str",addrOffset)){
          msg.getString(0, reaper.selectedTrack.insert[insertIndex].mixStr,VALUE_LENGTH); 
          if(selectedMode==INSERTS) {
            if (insertIndex<NUM_OF_CONTROLS and selectedInsertsPage==0) addToQueue(insertIndex+5);
            if (insertIndex>=NUM_OF_CONTROLS and selectedInsertsPage==1) addToQueue(insertIndex-NUM_OF_CONTROLS+5);
          }
          return;
        } 
      }

    }
    if (msg.match("/fx/*/fxparam",addrOffset)){
      showMessage(msg,addrOffset);
      return;
    }
      
   showMessage(msg,addrOffset);
   }
}

// parse insert/fx message and store incoming data in structs
void parseInsertMessage(OSCMessage &msg, int addrOffset ){
  // selected insert and track inserts
  if (msg.match("/fx")){
    char adresStr[48];    
    msg.getAddress(adresStr);
    int insertIndex = getAddresPartAsIndex(adresStr,2);
    if (msg.fullMatch("/fx/number/str")){
      // get selected insert number
      char number[2];
      msg.getString(0,number,2);
      reaper.selectedInsert.number=atoi(number);
      return;
    }
    if (msg.match("/fx/name")){
      msg.getString(0,reaper.selectedInsert.name,NAME_LENGTH); 
      identifyPlugin();
      if(selectedMode==PLUGIN) {
        addToQueue(31);
        addToQueue(30);
      }
//      functionTimer=0;
//      getParameterCount();
//      functionEndTime=functionTimer;
//      Serial.print("Count : ");
//      Serial.print(parameterCount);
//      Serial.print("\ttime : ");
//      Serial.println(functionEndTime);
      return;
    }  
    if (insertIndex>=0){
      if (msg.match("/fx/*/name")){
        msg.getString(0,reaper.selectedTrack.insert[insertIndex].name,NAME_LENGTH); 
          if(selectedMode==INSERTS) {
            if (insertIndex<NUM_OF_CONTROLS and selectedInsertsPage==0) addToQueue(insertIndex+13);
            if (insertIndex>=NUM_OF_CONTROLS and selectedInsertsPage==1) addToQueue(insertIndex-NUM_OF_CONTROLS+13);
          }
        return;
      }    
      if (msg.match("/fx/*/bypass")){
        reaper.selectedTrack.insert[insertIndex].bypass=int(msg.getFloat(0)); 
          if(selectedMode==INSERTS) {
            if (insertIndex<NUM_OF_CONTROLS and selectedInsertsPage==0) addToQueue(insertIndex+23);
            if (insertIndex>=NUM_OF_CONTROLS and selectedInsertsPage==1) addToQueue(insertIndex-NUM_OF_CONTROLS+23);
          }
        return;
      }
      if (msg.match("/fx/*/wetdry")){
        reaper.selectedTrack.insert[insertIndex].mix=msg.getFloat(0); 
//        Serial.print("Insert : ");
//        Serial.print(insertIndex);
//        Serial.print("\tMix : ");
//          Serial.println(reaper.selectedTrack.insert[insertIndex].mix);
        return;
      }
      if (msg.match("/fx/*/wetdry/str")){
        msg.getString(0, reaper.selectedTrack.insert[insertIndex].mixStr,VALUE_LENGTH); 
          if(selectedMode==INSERTS) {
            if (insertIndex<NUM_OF_CONTROLS and selectedInsertsPage==0) addToQueue(insertIndex+5);
            if (insertIndex>=NUM_OF_CONTROLS and selectedInsertsPage==1) addToQueue(insertIndex-NUM_OF_CONTROLS+5);
          }
        return;
      }
      if (msg.match("/fx/*/fxparam")){
        //showMessage(msg,addrOffset);
        return;
      }
    }

  }
}

// parse insert parameter message and store incoming data in struct
void parseInsertParameterMessage(OSCMessage &msg, int addrOffset ){
  // selected track insert parameters
  if (msg.match("/fxparam")){
    char adresStr[48];    
    msg.getAddress(adresStr);
    int index = getAddresPartAsIndex(adresStr,2);
    if (index>=0 and index< NUM_OF_PARAMETERS){
      if (msg.match("/fxparam/*/name")){
        msg.getString(0,reaper.insertParameter[index].name,NAME_LENGTH); 
        if (index-parameterOffset<4 and index-parameterOffset>=0){
          if(selectedMode==PLUGIN) addToQueue(index-parameterOffset+1);
        }
        return;
      }
      if (msg.fullMatch("/fxparam/*/value/str")){
        msg.getString(0,reaper.insertParameter[index].valueStr,VALUE_LENGTH); 
        if (index-parameterOffset<4 and index-parameterOffset>=0){
          if(selectedMode==PLUGIN) addToQueue(index-parameterOffset+5);
        }
        return;
      }
      if (msg.fullMatch("/fxparam/*/value")){
        if (index-parameterOffset<4 and index-parameterOffset>=0){
          if(processFeedback[index-parameterOffset]){
            reaper.insertParameter[index].value= msg.getFloat(0);
            if(selectedMode==PLUGIN) addToQueue(index-parameterOffset+9);
          }
        }
        else{
          reaper.insertParameter[index].value= msg.getFloat(0);
        }
        return;
      }
    }
  }
}

void centerString(char* string,unsigned int length){
  int   padCount=0;
  char  localString[length]={};

  if (strlen(string)<length-2) padCount=int((length-strlen(string))/2)+1; 
  for (int i=0;i<padCount;i++){
    strcat(localString," ");
  }
  strlcat(localString,string,length);
  strlcpy(centeredString,localString,length);
  
}

// CHECK PANEL BUTTONS & ENCODER SWITCHES
void checkButtons(){
  // read the button shift registers
  newPanelButtonStates = readButtonShiftRegisters();
  // reset the debouncing timer if the button states are not equal to the states in the previous scan
  if (newPanelButtonStates != debouncePanelButtonStates) debounceTimer = 0;
  // whatever the reading is at, it's been there for longer
  // than the debounce delay, so take it as the actual current state:
  if (debounceTimer >= debounceDelay) {
    // PANEL buttons
    if(newPanelButtonStates != currentPanelButtonStates){
      for(int i = 0; i < NUM_OF_BUTTONS; i++){
        if((newPanelButtonStates >> i) & 1){
            // if there is a LOW to HIGH transition (button pressed)  
            if (currentButtonState[i]==LOW){
              switch(i){
                case B_CURSOR_LEFT:
                  if (selectedMode==PLUGIN and currentButtonState[B_PLUGIN]==LOW) {
                    //sendOscMessage("/device/fxparam/bank/-",1.0);
                    if (reaper.selectedParameterBank>0){
                      reaper.selectedParameterBank--;
                      parameterOffset=reaper.selectedParameterBank*NUM_OF_CONTROLS;
                      for (int j=1;j<13;j++){
                        addToQueue(j);
                      }
                    }
                  }
                  else if (selectedMode==PLUGIN and currentButtonState[B_PLUGIN]==HIGH) {
                    sendOscMessage("/device/track/select",sourceTrackNumber);
                    selectMode(VOL_PAN);
                  }
                  else if (selectedMode==VOL_PAN) selectMode(SENDS);
                  else if (selectedMode==SENDS) selectMode(VOL_PAN);
                  else if (selectedMode==INSERTS){
                    if(selectedInsertsPage==1) selectedInsertsPage=0;
                    else selectedInsertsPage=1;
                    for (int j=13;j<17;j++){
                      addToQueue(j);
                    }
                    for (int j=23;j<27;j++){
                      addToQueue(j);
                    }
                  }
                  break;
                case B_CURSOR_RIGHT:
                  if (selectedMode==VOL_PAN) {
                    selectMode(SENDS);
                  }
                  else if (selectedMode==SENDS) {
                    selectMode(VOL_PAN);
                  }
                  else if (selectedMode==PLUGIN) {
                    //sendOscMessage("/device/fxparam/bank/+",1.0);
                    if (reaper.selectedParameterBank<NUM_OF_PARAMETERS/NUM_OF_CONTROLS){
                      reaper.selectedParameterBank++;
                      parameterOffset=reaper.selectedParameterBank*NUM_OF_CONTROLS;
                      for (int j=1;j<13;j++){
                        addToQueue(j);
                      }
                    }
                  }
                  else if (selectedMode==INSERTS){
                    if(selectedInsertsPage==0) selectedInsertsPage=1;
                    else selectedInsertsPage=0;
                    for (int j=13;j<17;j++){
                      addToQueue(j);
                    }
                    for (int j=23;j<27;j++){
                      addToQueue(j);
                    }
                  }
                  break;
                case B_CURSOR_UP:
                  clearPluginParameters();
                  if (selectedMode==PLUGIN) sendOscMessage("/device/fx/+",1.0);
                  if (selectedMode==VOL_PAN) sendOscMessage("/device/track/+",1.0);
                  if (selectedMode==SENDS) sendOscMessage("/device/track/+",1.0);
                  if (selectedMode==INSERTS) sendOscMessage("/device/track/+",1.0);
                  break;
                case B_CURSOR_DOWN:
                  clearPluginParameters();
                  if (selectedMode==PLUGIN) sendOscMessage("/device/fx/-",1.0);
                  if (selectedMode==VOL_PAN) sendOscMessage("/device/track/-",1.0);
                  if (selectedMode==SENDS) sendOscMessage("/device/track/-",1.0);
                  if (selectedMode==INSERTS) sendOscMessage("/device/track/-",1.0);
                  break;
                case B_TRACK:
                  if (selectedMode!=VOL_PAN) selectMode(VOL_PAN);
                  break;
                case B_MASTER:
                  if (selectedMode==PLUGIN and currentButtonState[B_PLUGIN]==HIGH){
                    scanPlugin();
                  }
                  else if (selectedMode!=MASTER) selectMode(MASTER);
                  break;
                case B_INSERTS:
                  if (selectedMode!=INSERTS) selectMode(INSERTS);
                  break;
                case B_PLUGIN:
                  if (selectedMode!=PLUGIN) selectMode(PLUGIN);
                  break;
              }
              currentButtonState[i]=HIGH;
              }
          }
        else{          
          if (currentButtonState[i]==HIGH){

            currentButtonState[i]=LOW;
            }
          }
      }
      currentPanelButtonStates = newPanelButtonStates;
    }
  }
  debouncePanelButtonStates = newPanelButtonStates;
  
  // read the encoder button shift registers
  newEncoderButtonStates = readEncoderShiftRegisters();
  // reset the debouncing timer if the button states are not equal to the states in the previous scan
  if (newEncoderButtonStates != debounceEncoderButtonStates) debounceEncoderTimer = 0;
  // whatever the reading is at, it's been there for longer
  // than the debounce delay, so take it as the actual current state:
  if (debounceEncoderTimer >= debounceDelay) {
    // if the state of one of the buttons has changed:
    if(newEncoderButtonStates != currentEncoderButtonStates){
      for(int i = 0; i < NUM_OF_ENCODER_BUTTONS; i++){
        if((newEncoderButtonStates >> i) & 1){
            // if there is a LOW to HIGH transition (button released)
            if (currentEncoderButtonState[i]==LOW){

              currentEncoderButtonState[i]=HIGH;
            }
          }
        else{          
          if (currentEncoderButtonState[i]==HIGH){
             // if there is a HIGH to LOW transition (button pressed)
            if (selectedMode==INSERTS){
              int selectedInsert=encoderButtonIndex[i]+selectedInsertsPage*NUM_OF_CONTROLS;
              char messagePath[48]={};
              reaper.selectedTrack.insert[selectedInsert].bypass=!reaper.selectedTrack.insert[selectedInsert].bypass;                
              strcat(messagePath,"/fx");
              strcat(messagePath,numToOSCAddress(selectedInsert+1));
              strcat(messagePath,"/bypass");
              sendOscMessage(messagePath,reaper.selectedTrack.insert[selectedInsert].bypass);
              break;
            }
            if (selectedMode==VOL_PAN){
              // get track number for send name, if found then select track and select plugin mode
              int sendTrack=0;
              for (int j=0;j<NUM_OF_TRACKS;j++){
                if (strcmp(reaper.selectedTrack.send[encoderButtonIndex[i]].name,reaper.track[j].name)==0){
                  sendTrack=reaper.track[j].number;
                  sourceTrackNumber=reaper.selectedTrack.number;
                  sendOscMessage("/device/track/select",sendTrack);
                  selectMode(PLUGIN);
                  break;
                }
              }
            }
            if (selectedMode==PLUGIN){
              if (currentButtonState[B_PLUGIN]==LOW){
                toggleFxParameter(encoderButtonIndex[i]);
              }
              else{
                pedalParameterAddress[0]=reaper.selectedTrack.number;
                pedalParameterAddress[1]=reaper.selectedInsert.number;
                pedalParameterAddress[2]=encoderButtonIndex[i]+parameterOffset+1;
              }
            }
            currentEncoderButtonState[i]=LOW;
            }
          }
      }
      currentEncoderButtonStates = newEncoderButtonStates;
    }
  }
  debounceEncoderButtonStates = newEncoderButtonStates;
}

// select controller mode
void selectMode(int mode){
  selectedMode=mode;
  // clear LEDs
  for (int i=0;i<NUM_OF_BUTTONS;i++) panelLedState[i]=LOW;
  // set LED for mode
  if (selectedMode==VOL_PAN or selectedMode==SENDS) panelLedState[4]=HIGH;
  if (selectedMode==MASTER) panelLedState[7]=HIGH;
  if (selectedMode==INSERTS) panelLedState[5]=HIGH;
  if (selectedMode==PLUGIN) panelLedState[6]=HIGH;
  switch(selectedMode){
    case VOL_PAN:
      deviderImageIndex[0]=1;
      deviderImageIndex[1]=2;
      deviderImageIndex[2]=2;
      break;
    case SENDS:
      deviderImageIndex[0]=1;
      deviderImageIndex[1]=1;
      deviderImageIndex[2]=1;
      break;
    case MASTER:
      deviderImageIndex[0]=2;
      deviderImageIndex[1]=1;
      deviderImageIndex[2]=1;
      break;
    case INSERTS:
      deviderImageIndex[0]=2;
      deviderImageIndex[1]=2;
      deviderImageIndex[2]=2;
      break;
    default:
      deviderImageIndex[0]=0;
      deviderImageIndex[1]=0;
      deviderImageIndex[2]=0;
      break;
  }
  refreshLeds=true;
  // update view
  addToQueue(31);
  addToQueue(30);
  for (int i=23;i<30;i++){
    addToQueue(i);
  }
  for (int i=16;i>0;i--){
    addToQueue(i);
  }
}


void getParameterCount(){
  parameterCount=0;
  for (int i=0;i<NUM_OF_PARAMETERS;i++){
    if (strlen(reaper.insertParameter[i].name)>1) {
      parameterCount++;
    } else break;
  }
}

void identifyPlugin(){
  pluginId=0;
  for (int i=0;i<NUM_OF_PLUGIN_MAPS;i++){
    if (strcmp(reaper.selectedInsert.name,pluginMap[i].name)==0){
      pluginId=i;
      break;
    }
  }
}

void clearPluginParameters(){
  for (int i=0;i<NUM_OF_PARAMETERS;i++){
    memset(&reaper.insertParameter[i].name[0], 0, sizeof( reaper.insertParameter[i].name));
    memset(&reaper.insertParameter[i].valueStr[0], 0, sizeof( reaper.insertParameter[i].valueStr));
    //reaper.insertParameter[i].name=0;
    //reaper.insertParameter[i].valueStr=0;
    reaper.insertParameter[i].value=0;
  }
}

void scanPlugin(){
  parameterCount=0;  
  for (int i=0;i<NUM_OF_PARAMETERS;i++){
    if (strlen(reaper.insertParameter[i].name)>1) {
      parameterCount++;
    } else break;
  }
  Serial.print("0,\"");
  Serial.print(reaper.selectedInsert.name);
  Serial.print("\",0,");
  Serial.print(parameterCount);
  Serial.println(",");
  Serial.println("\t{");
  for (int i=0;i<NUM_OF_PARAMETERS;i++){
    Serial.print("\t\t{\"");
    Serial.print(reaper.insertParameter[i].name);
    Serial.print("\"\t\t,");
    Serial.print(i+1);
    Serial.print(",0,1,100,\"   \",0,10");
    Serial.println("},");
  }
  Serial.println("\t},");
  Serial.println();
}

void checkEncoders(){
  // long for new encoder value
  long newEncoderValue = 0;
  for (int i=0;i<NUM_OF_ENCODERS;i++){
      char messagePath[48]={};
      newEncoderValue=encoder[i].read();
      if (newEncoderValue!=currentEncoderValue[i]) {
        // value encoders
        if (i<4){
          switch(selectedMode){
            case VOL_PAN:
              if (i<2){
                strcat(messagePath,"/track/send");
                strcat(messagePath,numToOSCAddress(i+1));
                strcat(messagePath,"/volume");
                reaper.selectedTrack.send[i].volume=constrain(reaper.selectedTrack.send[i].volume+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage(messagePath,reaper.selectedTrack.send[i].volume);
                addToQueue(i+9);
              }
              if (i==2){
                reaper.selectedTrack.pan=constrain(reaper.selectedTrack.pan+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage("/track/pan",reaper.selectedTrack.pan);
                addToQueue(11);
              }
              if (i==3){
                reaper.selectedTrack.volume=constrain(reaper.selectedTrack.volume+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage("/track/volume",reaper.selectedTrack.volume);
                addToQueue(12);
              }
              break;
            case SENDS:
                strcat(messagePath,"/track/send");
                strcat(messagePath,numToOSCAddress(i+1));
                strcat(messagePath,"/volume");
                reaper.selectedTrack.send[i].volume=constrain(reaper.selectedTrack.send[i].volume+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage(messagePath,reaper.selectedTrack.send[i].volume);
                addToQueue(i+9);
              break;
            case MASTER:
              // tempo
              if (i==0){
                reaper.tempo=constrain(reaper.tempo+(float(newEncoderValue-currentEncoderValue[i])*0.1),10,250);
                feedbackIgnoreTimer[i]=0;
                processFeedback[i]=false;
                sendOscMessage("/tempo/raw",reaper.tempo);
                addToQueue(20);
                addToQueue(9);
              }
//              if (i==1){
//                reaper.selectedTrack.hipassQ=constrain(reaper.selectedTrack.hipassQ+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
//                sendOscMessage("/fxeq/hipass/q",reaper.selectedTrack.hipassQ);
//                feedbackIgnoreTimer[i]=0;
//                processFeedback[i]=false;
//                addToQueue(10);
//              }
              // master pan
              if (i==2){
                reaper.masterPan=constrain(reaper.masterPan+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage("/master/pan",reaper.masterPan);
                addToQueue(11);
              }
              // master volume
              if (i==3){
                reaper.masterVolume=constrain(reaper.masterVolume+(float(newEncoderValue-currentEncoderValue[i])*encoderResolution),0.0,1.0);
                sendOscMessage("/master/volume",reaper.masterVolume);
                addToQueue(12);
              }
              break;
            case PLUGIN:    
              editFxParameter(i,newEncoderValue);
              break;
          }
          currentEncoderValue[i]=newEncoderValue; // update currentEncoderValue
        }
        // track select
        else if (i==4){
          if (abs(currentEncoderValue[i]-newEncoderValue) > 3){
            clearPluginParameters();
            if (newEncoderValue>currentEncoderValue[i]) sendOscMessage("/device/track/+",1.0);
            else sendOscMessage("/device/track/-",1.0);
            currentEncoderValue[i]=newEncoderValue; // update currentEncoderValue
          }
        }
        // jog shuttle
        else if (i==5){
          if (abs(currentEncoderValue[i]-newEncoderValue) > 3){
            if (newEncoderValue>currentEncoderValue[i]){
              // if cursor is at 1/00 move forward one measure, else move to start of next measure
              if (reaper.fraction==0 and reaper.beat==1) sendOscMessage("/action",41042);
              else  sendOscMessage("/action",41040);
            }
            else {
              // if cursor is at 1/00 move back one measure, else move to start of current measure
              if (reaper.fraction==0 and reaper.beat==1) sendOscMessage("/action",41043);
              else  sendOscMessage("/action",41041);
            }
            currentEncoderValue[i]=newEncoderValue; // update currentEncoderValue
          }
        }
      }
  }
}

void readPedal(){  
  char messagePath[48]={};
  pedal.update(); 
  if(pedal.hasChanged()) {
    int newPedalValue = pedal.getValue();
    if (abs(newPedalValue - pedalValue)>1){
      pedalValue=newPedalValue;
      strcat(messagePath,"/track");
      strcat(messagePath,numToOSCAddress(pedalParameterAddress[0]));
      strcat(messagePath,"/fx");
      strcat(messagePath,numToOSCAddress(pedalParameterAddress[1]));
      strcat(messagePath,"/fxparam");
      strcat(messagePath,numToOSCAddress(pedalParameterAddress[2]));
      strcat(messagePath,"/value");
//      Serial.print(messagePath);
//      Serial.print("\t\t");
//      Serial.println(constrain(float(pedalValue)/1010.0,0.0,1.0),3);
      sendOscMessage(messagePath,constrain(float(pedalValue)/1010.0,0.0,1.0));
    }
  }
}

// returns the array index for an address part, for use with track[], send[], insert[], parameter[]
int getAddresPartAsIndex(char* addres, int position){
  int number=0;   // function will return -1 if the position doesn't exist or doesn't contain an integer
  char* addresRest=addres;
  char* path=strtok_r(addresRest,"/",&addresRest);
  for (int i=1;i<position;i++){
    path=strtok_r(addresRest,"/",&addresRest);
  }
  number=atoi(path);
  return number-1;
}

//converts an integer to an osc address
char * numToOSCAddress( int index){
    static char s[10];
    int i = 9;
  
    s[i--]= '\0';
  do
    {
    s[i] = "0123456789"[index % 10];
                --i;
                index /= 10;
    }
    while(index && i);
    s[i] = '/';
    return &s[i];
}

// handlers for serial port, display ACK, NAK etc.
void serialEvent1() {
  while (displaySerial.available()) {
    byte incomingByte = displaySerial.read();
    switch(incomingByte){
      case GENIE_ACK :
        genieState=GENIE_LINK_IDLE;
        //Serial.print("Display: ACK ");
        break;
      case GENIE_NAK :   
        genieState=ERROR_NAK;
        Serial.println("Display: NAK ");
        break;    
      default: 
        Serial.print("Display received: ");
        Serial.println(incomingByte, HEX);
        break;
    }
  }
}


int  mapValue(float x, float in_min, float in_max, float out_min, float out_max) {
  return int((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min);
}

float  mapFloatValue(float x, float in_min, float in_max, float out_min, float out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

// edit FX parameter, called from encoder
void editFxParameter(int controlIndex, long newEncoderValue){
  int parameterIndex =controlIndex+parameterOffset;
  if (pluginId>0 and pluginMap[pluginId].parameter[parameterIndex].stepCount==1){
    return;
  }
  else {
    reaper.insertParameter[parameterIndex].value=constrain(reaper.insertParameter[parameterIndex].value+(float(newEncoderValue-currentEncoderValue[controlIndex])*encoderResolution),0.0,1.0);
    feedbackIgnoreTimer[controlIndex]=0;
    processFeedback[controlIndex]=false;
    sendFxParameter(parameterIndex);
    addToQueue(controlIndex+9);
  }
}

void toggleFxParameter (int controlIndex){
  int parameterIndex =controlIndex+parameterOffset;
  
  if (reaper.insertParameter[parameterIndex].value==0.0) {    
    reaper.insertParameter[parameterIndex].value=1.0;
    sendFxParameter(parameterIndex);
  }
  else if (reaper.insertParameter[parameterIndex].value==1.0) {
    reaper.insertParameter[parameterIndex].value=0.0;
    sendFxParameter(parameterIndex);
  }
}

// send FX parameter current value
void sendFxParameter(int index){
  char messagePath[48]={};          
  strcat(messagePath,"/fxparam");
  strcat(messagePath,numToOSCAddress(index+1));
  strcat(messagePath,"/value");
  sendOscMessage(messagePath,reaper.insertParameter[index].value);
}


//***** DISPLAY FUNCTIONS *****
void addToQueue(int objectIndex){
  // check if object is in queue
  if (objectQueue[objectIndex]==0){
    objectQueue[objectIndex]=queueLastNo;
    if (queueFirstIndex==0) queueFirstIndex=objectIndex;
    queueLastNo++;
  }
}

// set virtual 'knob' to FX parameter value
void displayParameterControl(int controlIndex){
  int parameterIndex =controlIndex+parameterOffset;
  // check if there is a mapping for this plugin
  if (pluginId>0){
    ParameterData *parameter=&pluginMap[pluginId].parameter[parameterIndex];
    // check parameter step count
    if (parameter->stepCount>0){
      genieWriteObject(GENIE_OBJ_USERIMAGES,controlIndex+8,mapValue(reaper.insertParameter[parameterIndex].value,0.0,1.0,parameter->imageOffset,parameter->imageOffset+parameter->stepCount));
    }
    else{
      genieWriteObject(GENIE_OBJ_USERIMAGES,controlIndex+8,mapValue(reaper.insertParameter[parameterIndex].value,0.0,1.0,parameter->imageOffset,parameter->imageOffset+300));
    }
  }
  else{
    genieWriteObject(GENIE_OBJ_USERIMAGES,controlIndex+8,mapValue(reaper.insertParameter[parameterIndex].value,0.0,1.0,10.0,310.0));
  }
}

// display FX parameter name
void displayParameterName(int controlIndex){
  int parameterIndex =controlIndex+parameterOffset;
  // check if there is a mapping for this parameter
  centerString(reaper.insertParameter[parameterIndex].name,NAME_LENGTH);
  genieWriteStr(controlIndex,centeredString);
}

// display FX parameter value
void displayParameterValue(int controlIndex){
  int parameterIndex =controlIndex+parameterOffset;
  char valueString[16]={};
  // check if there is a mapping for this parameter
  if (pluginId>0 and pluginMap[pluginId].parameter[parameterIndex].map){
    ParameterData *parameter=&pluginMap[pluginId].parameter[parameterIndex];
    dtostrf(mapFloatValue(reaper.insertParameter[parameterIndex].value,0.0,1.0,parameter->minValue,parameter->maxValue),1,parameter->precision,valueString);
    strcat(valueString,parameter->unit);
    centerString(valueString,VALUE_LENGTH);
    genieWriteStr(controlIndex+4,centeredString);
  }
  // display string, convert normalised values to 0-100 values
  else{
    strlcpy(valueString,reaper.insertParameter[parameterIndex].valueStr,VALUE_LENGTH);
    if (atof(valueString) and strlen(reaper.insertParameter[parameterIndex].valueStr)==6){
      double valueDouble=atof(valueString);
      if (valueDouble>=0.0 and valueDouble<=1.0){
        char converted[16]={};
        dtostrf(valueDouble*10.0,1,1,converted);
        centerString(converted,VALUE_LENGTH);
        genieWriteStr(controlIndex+4,centeredString);
      }else{
      centerString(reaper.insertParameter[parameterIndex].valueStr,VALUE_LENGTH);
      genieWriteStr(controlIndex+4,centeredString);
      }
    } 
    else if (strcmp(valueString,zeroString)==0){
      genieWriteStr(controlIndex+4,zeroLabel);
    }
    else{
      centerString(reaper.insertParameter[parameterIndex].valueStr,VALUE_LENGTH);
      genieWriteStr(controlIndex+4,centeredString);
    }
  }

}

// check display queue and update screen controls, texts and images
void updateDisplay(){
  char nameString[18]={};
  // handle queue for left display if state is idle
  
  // Screen objects
  // string 0-3       value labels under knobs
  // string 4-7       values for knobs
  // image  8-11      virtual knobs
  // string 8-11      labels above knobs
  // string 12        track number
  // string 15        track name
  // string 13        song position
  // string 14        tempo
  // gauge  0         VU L
  // gauge  1         VU R
  // image  0-3       Virtual button above knobs 0=blank, 1=green, 2=bypass, 3=mute
  // image  4-6       Devider lines between knobs 0=blank, 1=half, 2=full
  // image  7         Plugin icons
  if (genieState==GENIE_LINK_IDLE){
    // check if something in queue
    if (queueFirstIndex>0 ){
      // value labels
      if (queueFirstIndex>0 and queueFirstIndex<5){
        switch(selectedMode){
          case VOL_PAN:
            if (queueFirstIndex<3){
              centerString(reaper.selectedTrack.send[queueFirstIndex-1].name,NAME_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            if (queueFirstIndex==3) genieWriteStr(queueFirstIndex-1,emptyLabel);
            if (queueFirstIndex==4) genieWriteStr(queueFirstIndex-1,emptyLabel);
            break;
          case SENDS:
              centerString(reaper.selectedTrack.send[queueFirstIndex-1].name,NAME_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            break;
          case MASTER:
            genieWriteStr(queueFirstIndex-1,emptyLabel);
            break;
          case PLUGIN:
            displayParameterName(queueFirstIndex-1);
            break;
          default:
            genieWriteStr(queueFirstIndex-1,emptyLabel);
            break;
        }
      }
      // value strings
      else if (queueFirstIndex>4 and queueFirstIndex<9){
        switch(selectedMode){
          case VOL_PAN:
            if (queueFirstIndex<7){
              centerString(reaper.selectedTrack.send[queueFirstIndex-5].volumeStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            if (queueFirstIndex==7){
              centerString(reaper.selectedTrack.panStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            if (queueFirstIndex==8){
              centerString(reaper.selectedTrack.volumeStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            break;
          case SENDS:
              centerString(reaper.selectedTrack.send[queueFirstIndex-5].volumeStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            break;
          case MASTER:
            if (queueFirstIndex==5){
              genieWriteStr(queueFirstIndex-1,emptyLabel);
            }
            if (queueFirstIndex==6){
              genieWriteStr(queueFirstIndex-1,emptyLabel);
            }
            if (queueFirstIndex==7){
              centerString(reaper.masterPanStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            if (queueFirstIndex==8){
              centerString(reaper.masterVolumeStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            }
            break;
          case INSERTS:
              centerString(reaper.selectedTrack.insert[queueFirstIndex-5+selectedInsertsPage*NUM_OF_CONTROLS].mixStr,VALUE_LENGTH);
              genieWriteStr(queueFirstIndex-1,centeredString);
            break;
          case PLUGIN:
            displayParameterValue(queueFirstIndex-5);
            break;
          default:
            genieWriteStr(queueFirstIndex-1,emptyLabel);
            break;
        }
      }
      // knobs as userImages
      else if (queueFirstIndex>8 and queueFirstIndex<13){
        switch(selectedMode){
          case VOL_PAN:
            if (queueFirstIndex<11){
              genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-1,mapValue(reaper.selectedTrack.send[queueFirstIndex-9].volume,0.0,1.0,10.0,310.0));
            }
            if (queueFirstIndex==11) genieWriteObject(GENIE_OBJ_USERIMAGES,10,mapValue(reaper.selectedTrack.pan,0.0,1.0,10.0,310.0));
            if (queueFirstIndex==12) genieWriteObject(GENIE_OBJ_USERIMAGES,11,mapValue(reaper.selectedTrack.volume,0.0,1.0,10.0,310.0));
            break;
          case SENDS:
            genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-1,mapValue(reaper.selectedTrack.send[queueFirstIndex-9].volume,0.0,1.0,10.0,310.0));
            break;
          case MASTER:
            if (queueFirstIndex==9) genieWriteObject(GENIE_OBJ_USERIMAGES,8,constrain(mapValue(reaper.tempo,10.0,250.0,10.0,310.0),10,310));
            if (queueFirstIndex==10) genieWriteObject(GENIE_OBJ_USERIMAGES,9,0);
            if (queueFirstIndex==11) genieWriteObject(GENIE_OBJ_USERIMAGES,10,mapValue(reaper.masterPan,0.0,1.0,10.0,310.0));
            if (queueFirstIndex==12) genieWriteObject(GENIE_OBJ_USERIMAGES,11,mapValue(reaper.masterVolume,0.0,1.0,10.0,310.0));
            break;
          case PLUGIN:
            displayParameterControl(queueFirstIndex-9);
            break;
          default:
            genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-1,10);
            break;
        }
      }
      // labels above knobs
      else if (queueFirstIndex>12 and queueFirstIndex<17){
        switch(selectedMode){
          case VOL_PAN:
            if (queueFirstIndex==13) genieWriteStr(queueFirstIndex-5,sendsLabel);
            if (queueFirstIndex==14) genieWriteStr(queueFirstIndex-5,emptyLabel);
            if (queueFirstIndex==15) genieWriteStr(queueFirstIndex-5,panLabel);
            if (queueFirstIndex==16) genieWriteStr(queueFirstIndex-5,volumeLabel);
            break;
          case MASTER:
            if (queueFirstIndex==13) genieWriteStr(queueFirstIndex-5,tempoLabel);
            if (queueFirstIndex==14) genieWriteStr(queueFirstIndex-5,emptyLabel);
            if (queueFirstIndex==15) genieWriteStr(queueFirstIndex-5,panLabel);
            if (queueFirstIndex==16) genieWriteStr(queueFirstIndex-5,volumeLabel);
            break;
          case INSERTS:
              centerString(reaper.selectedTrack.insert[queueFirstIndex-13+selectedInsertsPage*NUM_OF_CONTROLS].name,NAME_LENGTH);
              genieWriteStr(queueFirstIndex-5,centeredString);
            break;
          case PLUGIN:
            genieWriteObject(GENIE_OBJ_STRINGS,queueFirstIndex-5,0);
            break;
          default:
            genieWriteObject(GENIE_OBJ_STRINGS,queueFirstIndex-5,0);
            break;
        }
      }
      // track number
      else if (queueFirstIndex==17){
        genieWriteStr(12,reaper.selectedTrack.number);
      }
      // track name
      else if (queueFirstIndex==18){
        strlcpy(nameString,reaper.selectedTrack.name,18);
        genieWriteStr(15,nameString);
      }
      // song position
      else if (queueFirstIndex==19){
        char measureBeatString[7]={};
        char measurePart[4];
        utoa(reaper.measure,measurePart,10);
        char beatPart[2];
        utoa(reaper.beat,beatPart,10);
        strcat(measureBeatString,measurePart);
        strcat(measureBeatString,".");
        strcat(measureBeatString,beatPart);
        genieWriteStr(13,measureBeatString);
      }
      // tempo              
      else if (queueFirstIndex==20){
        char tempoStr[6]={};
        dtostrf(reaper.tempo,1,1,tempoStr);
        genieWriteStr(14,tempoStr);
      }
      // VU L
      else if (queueFirstIndex==21){
        if (selectedMode==MASTER) genieWriteObject(GENIE_OBJ_GAUGE,0,int(reaper.masterVuL*100.0));
        else genieWriteObject(GENIE_OBJ_GAUGE,0,int(reaper.trackVuL*100.0));
      }
      // VU R
      else if (queueFirstIndex==22){
        if (selectedMode==MASTER) genieWriteObject(GENIE_OBJ_GAUGE,1,int(reaper.masterVuR*100.0));
        else genieWriteObject(GENIE_OBJ_GAUGE,1,int(reaper.trackVuR*100.0));
      }
      // virtual buttons (images) above knobs
      else if (queueFirstIndex>22 and queueFirstIndex<27){
        if (selectedMode==INSERTS){
          if (strlen(reaper.selectedTrack.insert[queueFirstIndex-23+selectedInsertsPage*NUM_OF_CONTROLS].name)>1){
            if (reaper.selectedTrack.insert[queueFirstIndex-23+selectedInsertsPage*NUM_OF_CONTROLS].bypass) genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-23,1);
            else genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-23,2);
          } else genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-23,0);
        } else genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-23,0);

      }
      // devider images between knobs 
      else if (queueFirstIndex>26 and queueFirstIndex<30){
        genieWriteObject(GENIE_OBJ_USERIMAGES,queueFirstIndex-23,deviderImageIndex[queueFirstIndex-27]);
      }
      // Plugin icon
      else if (queueFirstIndex==30){
        int iconIndex=0;
        switch(selectedMode){
          case PLUGIN:
           for (int i=0;i<NUM_OF_ICON_MAPS;i++){
              if (strcmp(reaper.selectedInsert.name,iconMap[i].name)==0){
                iconIndex=iconMap[i].iconIndex;
                break;
              }
            }
            genieWriteObject(GENIE_OBJ_USERIMAGES,7,iconIndex);
            break;
          default:
            genieWriteObject(GENIE_OBJ_USERIMAGES,7,0);
            break;
        }
      }
      // Big label
      else if (queueFirstIndex==31){
        switch(selectedMode){
          case VOL_PAN:
            genieWriteObject(GENIE_OBJ_STRINGS,16,1);
            break;
          case SENDS:
            genieWriteObject(GENIE_OBJ_STRINGS,16,2);
            break;
          case MASTER:
            genieWriteObject(GENIE_OBJ_STRINGS,16,3);
            break;
          case INSERTS:
            genieWriteObject(GENIE_OBJ_STRINGS,16,4);
            break;
          case PLUGIN:
            genieWriteStr(16,reaper.selectedInsert.name);
            break;
          default:
            genieWriteObject(GENIE_OBJ_STRINGS,16,0);
            break;
        }
      }
      
      queueFirstIndex=0; // reset and determine again
      for (int i=1;i<=NUM_OF_SCREEN_OBJECTS;i++){
        if (objectQueue[i] > 0)  objectQueue[i]--;
        if (objectQueue[i] == 1) queueFirstIndex=i;
      }
      queueLastNo--;
    }
  }
}

//***** 4D Genie functions copied from library *****
void genieWriteObject (uint16_t object, uint16_t index, uint16_t data) {
    uint16_t msb, lsb ;
    uint8_t checksum ;
    lsb = lowByte(data);
    msb = highByte(data);
    displaySerial.write(GENIE_WRITE_OBJ) ;
    checksum  = GENIE_WRITE_OBJ ;
    displaySerial.write(object) ;
    checksum ^= object ;
    displaySerial.write(index) ;
    checksum ^= index ;
    displaySerial.write(msb) ;
    checksum ^= msb;
    displaySerial.write(lsb) ;
    checksum ^= lsb;
    displaySerial.write(checksum) ;
    genieState=GENIE_LINK_WFAN;

}
void genieWriteStr (uint16_t index, char *string) {
    char *p;
    unsigned int checksum;
    int len = strlen (string);
    displaySerial.write(GENIE_WRITE_STR);
    checksum  = GENIE_WRITE_STR;
    displaySerial.write(index);
    checksum ^= index;
    displaySerial.write((unsigned char)len);
    checksum ^= len;
    for (p = string ; *p ; ++p) {
        displaySerial.write(*p);
        checksum ^= *p;
    }
    displaySerial.write(checksum);
    genieState=GENIE_LINK_WFAN;
}
void genieWriteStr(uint16_t index, const String &s){
  //s.c_str(), s.length()
  int len = s.length();
  char arr[len + 1];
  s.toCharArray(arr,len + 1);
  genieWriteStr(index, arr);
}
void genieWriteContrast (uint16_t value) {
    unsigned int checksum ;
    displaySerial.write(GENIE_WRITE_CONTRAST) ;
    checksum  = GENIE_WRITE_CONTRAST ;
    displaySerial.write(value) ;
    checksum ^= value ;
    displaySerial.write(checksum) ;
    genieState=GENIE_LINK_WFAN;
}

// read shift registers (scan buttons)
unsigned int readButtonShiftRegisters(){
  int bitVal;
  byte bytesVal = 0;
  // Trigger a parallel Load to latch the state of the data lines
  digitalWrite(PLOAD_PIN, LOW);
  digitalWrite(PLOAD_PIN, HIGH);
  // Loop to read each bit value from the serial out line of the SN74HC165N.
  for(int i = 0; i < NUM_OF_BUTTONS; i++)
  {
      bitVal = digitalRead(DATA_PIN);
      // Set the corresponding bit in bytesVal.
      bytesVal |= (bitVal << (NUM_OF_BUTTONS-1 - i));
      // Pulse the Clock (rising edge shifts the next bit).
      digitalWrite(CLOCK_PIN, HIGH);
      digitalWrite(CLOCK_PIN, LOW);
  }
  return(bytesVal);
}

unsigned int readEncoderShiftRegisters(){
  int bitVal;
  unsigned int bytesVal = 0;
  // Trigger a parallel Load to latch the state of the data lines
  digitalWrite(PLOAD_PIN_ENCODERS, LOW);
  digitalWrite(PLOAD_PIN_ENCODERS, HIGH);
  // Loop to read each bit value from the serial out line of the SN74HC165N.
  for(int i = 0; i < NUM_OF_ENCODER_BUTTONS; i++)
  {
      bitVal = digitalRead(DATA_PIN_ENCODERS);
      // Set the corresponding bit in bytesVal.
      bytesVal |= (bitVal << (NUM_OF_ENCODER_BUTTONS-1 - i));
      // Pulse the Clock (rising edge shifts the next bit).
      digitalWrite(CLOCK_PIN_ENCODERS, HIGH);
      digitalWrite(CLOCK_PIN_ENCODERS, LOW);
  }
  return(bytesVal);
}

// write to shift registers (update leds)
void updateLeds(){
  //ground latchPin and hold low for as long as you are transmitting
  digitalWrite(LATCH_PIN, 0);
  //clear everything out just in case to
  //prepare shift register for bit shifting
  digitalWrite(LED_DATA_PIN, 0);
  digitalWrite(LED_CLOCK_PIN, 0);
  // step through  panelLedState array
  for (int i=NUM_OF_BUTTONS-1; i>=0; i--)  {
    digitalWrite(LED_CLOCK_PIN, 0);
    //Sets the pin to HIGH or LOW depending on ledState
    digitalWrite(LED_DATA_PIN, panelLedState[i]);
    //register shifts bits on upstroke of clock pin  
    digitalWrite(LED_CLOCK_PIN, 1);
    //zero the data pin after shift to prevent bleed through
    digitalWrite(LED_DATA_PIN, 0);
  }
  //stop shifting
  digitalWrite(LED_CLOCK_PIN, 0);
  //return the latch pin high to signal chip that it 
  //no longer needs to listen for information
  digitalWrite(LATCH_PIN, 1);
  refreshLeds=false;
}

// print track data to serial port
void printTrackReport(){
  Serial.println("SELECTED TRACK PARAMETERS");
  Serial.print("\tName\t\t");
  Serial.println(reaper.selectedTrack.name);
  Serial.print("\tNumber\t\t");
  Serial.println(reaper.selectedTrack.number);
  Serial.print("\tVolume\t\t");
  Serial.println(reaper.selectedTrack.volumeStr);
  Serial.print("\tPan\t\t");
  Serial.println(reaper.selectedTrack.panStr);
  Serial.print("\tRecord\t\t");
  Serial.println(reaper.selectedTrack.rec);
  Serial.print("\tMute\t\t");
  Serial.println(reaper.selectedTrack.mute);
  Serial.print("\tSolo\t\t");
  Serial.println(reaper.selectedTrack.solo);
  Serial.print("\tMonitor\t\t");
  Serial.println(reaper.selectedTrack.monitor);
  Serial.println();
  for (int i=0;i<NUM_OF_SENDS;i++){
    Serial.print("\tSend ");
    Serial.print(i+1);
    Serial.print("\t\t");
    Serial.print(reaper.selectedTrack.send[i].name);
    Serial.print("\t\t");
    Serial.print(reaper.selectedTrack.send[i].volumeStr);
    Serial.print("\t\t");
    Serial.print(reaper.selectedTrack.send[i].panStr);
    Serial.println();
  }
  Serial.println();
  for (int i=0;i<NUM_OF_INSERTS;i++){
    Serial.print("\tInsert ");
    Serial.print(i+1);
    Serial.print("\t");
    Serial.print(reaper.selectedTrack.insert[i].bypass);
    Serial.print("\t");
    Serial.print(reaper.selectedTrack.insert[i].mix);
    Serial.print("\t");
    Serial.print(reaper.selectedTrack.insert[i].mixStr);
    Serial.print("\t");
    Serial.print(reaper.selectedTrack.insert[i].name);
    Serial.print("--");
    Serial.println();
  }
  Serial.println();
  Serial.println("SELECTED INSERT PARAMETERS");
  for (int i=0;i<NUM_OF_PARAMETERS;i++){
    Serial.print("\t");
    Serial.print(reaper.insertParameter[i].name);
    Serial.print("\t\t");
    Serial.print(reaper.insertParameter[i].valueStr);
    Serial.println();
  }
  Serial.println();
  Serial.println("TRACK OVERVIEW");
  for (int i=0;i<NUM_OF_TRACKS;i++){
    Serial.print(reaper.track[i].name);
    Serial.print("\t\t");
    Serial.print("\t");
    Serial.print(reaper.track[i].number);
    Serial.print("\t");
    Serial.print(reaper.track[i].volumeStr);
    Serial.print("\t\t");
    Serial.print(reaper.track[i].panStr);
    Serial.print("\t");
    Serial.print(reaper.track[i].insert[0].name);
    Serial.print("\t\t");
    Serial.println(reaper.track[i].send[0].name);
    
  }
}
